"""Fixpoint iterator over a CfG

- Transfer is a base class that must be instantiated for
each static analysis.
- fixpoint_iteration computes a fixpoint over a cfg for the given
transfer instance
"""
from cfg import *
from typing import Protocol

class Transfer[S](Protocol):
    """lists the functions that must be implemented for a static analysis over Lattice S"""
    def bottom(self) -> S:
        """the empty state, associated to unvisited nodes"""
        ...
    def init_state(self) -> S:
        """the state associated to the initial node"""
        ...
    def join(self,s1:S,s2:S) -> S:
        """join of two states when merging branches"""
        ...
    def included(self,s1: S,s2: S) -> bool:
        """return True when the first state is included in the second"""
        ...
    def tr_skip(self,s: S) -> S:
        """transfer state across a Skip transition

        Almost always transfer the state unchanged
        """
        return s
    def tr_set(self,s: S,v: str,e: ArithExpr) -> S:
        """transfer state across an assignment"""
        ...
    def tr_test(self,s: S,c: BoolExpr) -> S:
        """transfer state across a test transition"""
        ...
    def tr_err(self,s: S,e: Expr) -> S:
        """transfer state across a transition to error state"""
        ...

def fixpoint_iteration[T](transfer: Transfer[T], cfg: Cfg) -> dict[Node,T]:
    """computes the fixpoint for the given transfer functions over the given CfG
 
   output maps each node to the computed state
    """
    raise NotImplementedError
